# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

from base64 import b64decode
import datetime
import logging
import os
import psutil
import re
import time

import cmf.log_search
from cmf.throttling_logger import ThrottlingLogger
from cmf.url_util import urlopen_with_timeout

import clusterstats.common.chain as chain
from clusterstats.log.streaming.file import LogFile
from clusterstats.log.streaming.event import LogEvent
from clusterstats.log.streaming.event_reader import LogFileEventReader
from clusterstats.log.streaming.event_streamer import LogEventStreamer
from clusterstats.log.streaming.file_filter import DirectoryFilter
from clusterstats.log.streaming.file_filter import FilePatternFilter

LOG = logging.getLogger('ImpaladQueryMonitor')
THROTTLED_LOG = ThrottlingLogger(LOG)

# Used to monitor our resource consumption.
_PROCESS = psutil.Process(os.getpid())

def _safe_get_io_counters():
  try:
    return _PROCESS.get_io_counters()
  except psutil.NoSuchProcess:
    # We don't support I/O metrics on linux machines with kernels that don't
    # support I/O Stats (that don't have /proc/<pid>/io)
    return psutil.ntuple_io(0,0,0,0)

# We implement our own TimeRangeFilter that's identical in how it filters to the
# custerstats one. The only real difference is that we keep track of the time
# of the latest event time within the requested filter range.
class _TimeRangeFilter(object):
  """Filter out log files that don't have log entries within the given
  time range (represented by the start_date_time and end_date_time passed
  to the constructor). It does this by peaking at the first and last log
  event in the file, and assuming that log events are in time increasing
  order."""

  def __init__(self, start_date_time, end_date_time, event_reader):
    self.__start_date_time = start_date_time
    self.__end_date_time = end_date_time
    self.__event_reader = event_reader

    self.latest_event_datetime = None

  def __call__(self, context):
    new_file_list = []
    for f in context["file_list"]:
      file_handle = f.open_log_file()
      try:
        self.__event_reader.initialize(file_handle, f.get_path())
        evt1 = self.__event_reader.get_first_event()
        if evt1 is not None and evt1.get_datetime() <= self.__end_date_time:
          evt2 = self.__event_reader.get_last_event()
          if evt2 is not None and evt2.get_datetime() >= self.__start_date_time:
            self.__set_start_offset(f, evt1.get_datetime())
            self.__set_end_offset(f, evt2.get_datetime())
            new_file_list.append(f)
      finally:
        file_handle.close()

    context["file_list"] = new_file_list
    return True

  def __set_start_offset(self, f, first_event_datetime):
    if first_event_datetime < self.__start_date_time:
      event = self.__event_reader.find_nearest_event_datetime(self.__start_date_time)
      f.set_start_offset(event.get_offset())

  # This function contains the only changes to the clusterstats TimeRangeFilter.
  def __set_end_offset(self, f, last_event_datetime):
    end_datetime = last_event_datetime

    if last_event_datetime > self.__end_date_time:
      event = self.__event_reader.find_nearest_event_datetime(self.__end_date_time)
      f.set_end_offset(event.get_offset())
      end_datetime = event.get_datetime()

    if (self.latest_event_datetime is None or
        end_datetime > self.latest_event_datetime):
      self.latest_event_datetime = end_datetime

# We extend LogEvent to explicitly track the query ID and profile we parse out
# below. We also simply hard code some fields that do not make sense for our
# file.
class _QueryLogEvent(LogEvent):
  def __init__(self,
               file_path,
               offset,
               end_offset,
               date_time,
               query_id,
               profile):
    LogEvent.__init__(self,
                      file_path,
                      offset,
                      end_offset,
                      date_time,
                      "INFO",
                      "QUERYLOG",
                      "<query>")
    self.query_id = query_id
    self.profile = profile

# Every line in the file is as follows:
#
# 136924029461 63488804fc4451a8:ab2d3f8e4134a18e eJyFlU1...
#
# The format here is the timestamp in millis since the epoch, then the query ID,
# then a base64 encoded, zlib compressed compact, binary thrift profile.
class _QueryLogEventParser(object):

    def __init__(self):
      timestamp_format = "\d+"
      query_id_format = "\S+"
      profile_format = ".+\\n"
      regex_string = "(?P<timestamp>" + timestamp_format + \
          ")\s+(?P<query_id>" + query_id_format + \
          ")\s+(?P<profile>" + profile_format + ")"
      self.__regex = re.compile(regex_string)

    def parse_event(self, line):
      m = self.__regex.match(line.get_content())
      if m is not None:
        return self.__construct_event(
          line.get_file_path(),
          line.get_offset(),
          line.get_offset() + len(line.get_content()),
          m)
      else:
        return None

    def __construct_event(self, file_path, offset, end_offset, m):
      dt = datetime.datetime.utcfromtimestamp(long(m.group("timestamp")) / 1000.0)
      return _QueryLogEvent(file_path,
                            offset,
                            end_offset,
                            dt,
                            m.group("query_id"),
                            m.group("profile"))

# If we request a profile for which there is no information we get a message
# like this: "Query id 323c59bd14a544f5:a8329df97b197e9a not found."
_PROFILE_NOT_FOUND_PREFIX = "Query id"

# Scans the logical log file at the input path for completed query entries
# between the start and end times, starting with the input query ID . This is
# based on our log search code, so it will handle multiple log files, rotation,
# etc. This function returns a tuple of the following:
# 
# (next start datetime to use to continue the search,
#  list of (query ID, thrift profile) tuples.)
#
# log_path: path to the log file to search
# start_datetime: start of the search time window
# end_datetime: end of the search time window
# last_reported_query_id: should be specified when streaming to avoid double
#  reporting of queries, this is the query id of the last full query seen by
#  the previous stream searches
# limit: no more than this number of completed queries will be returned
# stats: statistics object for tracking various things
#
# Profiles are returned in zlib compressed, binary compact thrift format.
def _get_completed_queries(log_path,
                           start_datetime,
                           end_datetime,
                           last_reported_query_id,
                           limit,
                           stats):
  log_directory, log_filename = os.path.split(log_path)
  if not os.path.isdir(log_directory):
    return None, []

  reader = LogFileEventReader(event_parser=_QueryLogEventParser())
  time_range_filter = _TimeRangeFilter(start_datetime, end_datetime, reader)
  filters = chain.build_chain([DirectoryFilter(),
                               FilePatternFilter(log_filename + '*'),
                               time_range_filter])
  streamer = LogEventStreamer(log_directory,
                              reader,
                              file_filter=filters)

  completed_queries = []
  # If we hit the limit on the number of queries we'll report, then we want to
  # start at that event next time.
  limit_event_datetime = None

  for event in streamer.stream():
    if len(completed_queries) >= limit:
      stats.increment('total_completed_query_limit_reached')
      LOG.debug("Hit completed query limit.")
      limit_event_datetime = event.get_datetime()
      break

    query_id = event.query_id
    profile = event.profile

    # If we run into the last reported query ID, then clear any accumulated
    # completed queries.
    if query_id == last_reported_query_id:
      completed_queries = []
      continue

    try:
      completed_queries.append((query_id, b64decode(profile)))
    except Exception, ex:
      LOG.exception("Error parsing query with id '%s': '%s'" % 
                    (query_id, profile))
      stats.increment('total_completed_query_parse_exceptions')

  # We need to do a little bookkeeping now. We will return the start time that
  # the next query should use. Note that this is not determined by the end time
  # we used above. It is determined by the time of the last event our log search
  # actually saw, or the time of the event on which we hit our return limit. If
  # we didn't stream any events, then we'll say to just keep using the passed in
  # start time.
  if limit_event_datetime is not None:
    next_start_datetime = limit_event_datetime
  elif time_range_filter.latest_event_datetime is not None:
    next_start_datetime = time_range_filter.latest_event_datetime
  else:
    next_start_datetime = start_datetime

  return next_start_datetime, completed_queries

# Parses the data returned by the executing query ID servlet, returning a list of
# query IDs. We expect the servlet to return plain text with query IDs separated
# by a single space.
def _parse_executing_query_ids(opened_url):
  query_ids_string = opened_url.read()
  if len(query_ids_string) == 0:
    return []
  query_ids = query_ids_string.split('\n')
  # An extra end line creates an empty query_id so let's remove that
  if len(query_ids) > 0 and query_ids[-1] == "":
    query_ids = query_ids[:-1]
  return query_ids

# Parses the data returned by the executing query profile servlet, returning an
# encoded zlib compressed, binary compact thrift query profile or None if one
# could not be parsed.
def _parse_executing_query_profile(opened_url):
  profile = opened_url.read()
  if len(profile) == 0 or profile.startswith(_PROFILE_NOT_FOUND_PREFIX):
    return None
  return b64decode(profile)

class _Stats(object):
  def __init__(self):
    # The following counters offer an approximation of the cost of the query
    # monitor. They are approximate because they rely on psutil's process wide
    # CPU and IO counters. As such, it's possible for them to get thrown off by
    # some concurrent expensive operation by the agent. Still, they are better
    # than nothing.
    self.total_get_completed_query_calls = 0
    self.total_wallclock_time_seconds = 0.0
    self.total_user_cpu_time_seconds = 0.0
    self.total_system_cpu_time_seconds = 0.0
    # This is basically rchar from /proc/io.
    self.total_logical_bytes_read = 0
    # This is basically read_butes from /proc/io.
    self.total_storage_bytes_read = 0
    # This is the total size in bytes of all the profile data we've returned.
    self.total_completed_queries_size_bytes = 0
    self.total_executing_queries_size_bytes = 0
    # The following are lifetime counts of the number of queries and ids
    # retrieved by this query monitor.
    self.total_completed_queries = 0
    self.total_executing_queries = 0
    self.total_executing_queries_ids = 0
    # The following are lifetime error counts for this query monitor.
    self.total_slow_completed_query_warnings = 0
    self.total_unexpected_start_time_advances = 0
    self.total_get_executing_query_ids_exceptions = 0
    self.total_get_executing_query_profile_exceptions = 0
    self.total_completed_query_limit_reached = 0
    self.total_completed_query_parse_exceptions = 0

  def increment(self, attribute, delta=1):
    v = getattr(self, attribute)
    setattr(self, attribute, v + delta)

  def to_dict(self):
    return dict({
        'total_get_completed_query_calls':
          self.total_get_completed_query_calls,
        'completed_query_monitor_total_wallclock_time_seconds':
          self.total_wallclock_time_seconds,
        'completed_query_monitor_total_user_cpu_time_seconds':
          self.total_user_cpu_time_seconds,
        'completed_query_monitor_total_system_cpu_time_seconds':
          self.total_system_cpu_time_seconds,
        'completed_query_monitor_total_logical_bytes_read':
          self.total_logical_bytes_read,
        'completed_query_monitor_total_storage_bytes_read':
          self.total_storage_bytes_read,
        'total_completed_queries':
          self.total_completed_queries,
        'total_executing_queries':
          self.total_executing_queries,
        'total_executing_queries_ids':
          self.total_executing_queries_ids,
        'total_slow_completed_query_warnings':
          self.total_slow_completed_query_warnings,
        'total_unexpected_start_time_advances':
          self.total_unexpected_start_time_advances,
        'total_get_executing_query_ids_exceptions':
          self.total_get_executing_query_ids_exceptions,
        'total_get_executing_query_profile_exceptions':
          self.total_get_executing_query_profile_exceptions,
        'total_completed_query_limit_reached':
          self.total_completed_query_limit_reached,
        'total_completed_query_parse_exceptions':
          self.total_completed_query_parse_exceptions,
        'total_completed_queries_size_bytes':
          self.total_completed_queries_size_bytes,
        'total_executing_queries_size_bytes':
          self.total_executing_queries_size_bytes
        })

# This class exists to allow for easier testing of the overall Impala Daemon
# monitor. It also handles tracking various query monitoring statistics.
class QueryMonitor(object):
  # We take an optional urlopen_with_timeout function as a parameter to help with
  # testing. If passed it, this function can be used to mock out urllib requests.
  #
  # The completed_query_slow_log_threshold_seconds specifies a threshold for the
  # wallclock time that a call to get_completed_queries can take before we log
  # a warning that it is taking to long.
  def __init__(self,
               urlopen_with_timeout=urlopen_with_timeout,
               completed_query_slow_log_threshold_seconds=1.0):
    self._urlopen_with_timeout = urlopen_with_timeout
    self._completed_query_slow_log_threshold_seconds = \
        completed_query_slow_log_threshold_seconds
    self._stats = _Stats()

  def get_completed_queries(self,
                            log_path,
                            start_datetime,
                            end_datetime,
                            last_reported_query_id,
                            limit):
    start = time.time()
    start_cpu = _PROCESS.get_cpu_times()
    start_io = _safe_get_io_counters()

    next_start_datetime, completed_query_profiles = \
        _get_completed_queries(log_path,
                               start_datetime,
                               end_datetime,
                               last_reported_query_id,
                               limit,
                               self._stats)

    completed_query_count = 0
    if completed_query_profiles is not None:
      completed_query_count = len(completed_query_profiles)
      self._stats.increment('total_completed_queries', completed_query_count)
      for p in completed_query_profiles:
        self._stats.increment('total_completed_queries_size_bytes', len(p))

    end = time.time()
    end_cpu = _PROCESS.get_cpu_times()
    end_io = _safe_get_io_counters()    

    delta_wallclock = end - start
    delta_cpu_user = end_cpu.user - start_cpu.user
    delta_cpu_system = end_cpu.system - start_cpu.system
    delta_logical_bytes = end_io.read_count - start_io.read_count
    delta_storage_bytes = end_io.read_bytes - start_io.read_bytes

    if delta_wallclock > self._completed_query_slow_log_threshold_seconds:
      self._stats.increment('total_slow_completed_query_warnings')
      LOG.warn(("Call to get completed Impala queries too longer than %f " 
                "seconds: wallclock: %f user cpu %f system CPU %f "
                "logical bytes read: %d storage bytes read: %d queries "
                "fetched: %d. Counters reflect work by entire agent, not just "
                "the query monitor.") % 
               (self._completed_query_slow_log_threshold_seconds,
                delta_wallclock,
                delta_cpu_user,
                delta_cpu_system,
                delta_logical_bytes,
                delta_storage_bytes,
                completed_query_count))

    self._stats.increment('total_get_completed_query_calls')
    self._stats.increment('total_wallclock_time_seconds', delta_wallclock)
    self._stats.increment('total_user_cpu_time_seconds', delta_cpu_user)
    self._stats.increment('total_system_cpu_time_seconds', delta_cpu_system)
    self._stats.increment('total_logical_bytes_read', delta_logical_bytes)
    self._stats.increment('total_storage_bytes_read', delta_storage_bytes)

    return next_start_datetime, completed_query_profiles

  # Gets a list of IDs of currently executing queries by hitting the Impala Daemon
  # web server at the input url.
  def get_executing_query_ids(self, url, timeout, username=None, password=None):
    if url is None:
      LOG.error("A url parameter is required.")
      return []

    try:
      opened_url = self._urlopen_with_timeout(url,
                                              timeout=timeout,
                                              username=username,
                                              password=password)
      query_ids = _parse_executing_query_ids(opened_url)
    except Exception, ex:
      LOG.exception("Error fetching executing query ids at '%s'" % (url,))
      self._stats.increment('total_get_executing_query_ids_exceptions')
      return []

    if query_ids is not None:
      self._stats.increment('total_executing_queries_ids', len(query_ids))

    return query_ids

  # Gets a base64 encoded compact thrift format runtime profile from the Impala
  # Daemon web server at the input url for the input query id. This will return
  # None on a timeout, an error or if the profile could not be found.
  def get_executing_query_profile(self, url, query_id, timeout, username=None,
                                  password=None):
    if url is None:
      LOG.error("A url parameter is required.")
      return None
    if query_id is None:
      LOG.error("A query_id parameter is required.")
      return None
    url_with_parms = "%s?query_id=%s" % (url, query_id)
    try:
      opened_url = self._urlopen_with_timeout(url_with_parms,
                                              timeout=timeout,
                                              username=username,
                                              password=password)
      profile = _parse_executing_query_profile(opened_url)
    except Exception, ex:
      LOG.exception("Error fetching executing query profile at '%s'" %
                    (url_with_parms,))
      self._stats.increment('total_get_executing_query_profile_exceptions')
      return None

    if profile is not None:
      self._stats.increment('total_executing_queries')
      self._stats.increment('total_executing_queries_size_bytes', len(profile))

    return profile

  def get_stats(self):
    return self._stats.to_dict()

# This extension of the QueryMonitor class handles the bookkeeping necessary to
# monitor the stream of queries in an Impala Daemon log file.
class QueryStreamMonitor(QueryMonitor):
  def __init__(self,
               start_datetime=None,
               urlopen_with_timeout=urlopen_with_timeout,
               completed_query_slow_log_threshold_seconds=1.0):
    QueryMonitor.__init__(self,
                          urlopen_with_timeout,
                          completed_query_slow_log_threshold_seconds)
    if start_datetime is None:
      start_datetime = datetime.datetime.today()

    # These members tracks the start time we will use in our next check of the log
    # for completed impala queries.
    self._completed_queries_monitoring_start_datetime = start_datetime
    self._completed_queries_monitoring_last_reported_query_id = None

  def get_completed_queries(self,
                            log_file,
                            end_datetime=None,
                            completed_query_report_limit=1000):

    if end_datetime is None:
      end_datetime = datetime.datetime.today()

    next_start_datetime, completed_queries = QueryMonitor.get_completed_queries(
      self,
      log_file,
      self._completed_queries_monitoring_start_datetime,
      end_datetime,
      self._completed_queries_monitoring_last_reported_query_id,
      completed_query_report_limit)

    if next_start_datetime is None:
      # If we got no time to use for our next monitoring call, just go ahead and
      # use the current time as the new start time. This should only happen if
      # there's funny business going on.
      THROTTLED_LOG.warning("Call to get completed queries did not return a "
                            "new start time. Advancing query monitoring "
                            "period.")
      self._completed_queries_monitoring_start_datetime = \
          datetime.datetime.today()
      self._stats.increment('total_unexpected_start_time_advances')
    else:
      self._completed_queries_monitoring_start_datetime = next_start_datetime

    if len(completed_queries) > 0:
      self._completed_queries_monitoring_last_reported_query_id = \
          completed_queries[-1][0]

    return completed_queries

  def get_stats(self):
    stats = QueryMonitor.get_stats(self)
    stats['completed_queries_monitoring_start_datetime'] = \
        str(self._completed_queries_monitoring_start_datetime)
    stats['completed_queries_monitoring_last_reported_query_id'] = \
        self._completed_queries_monitoring_last_reported_query_id
    return stats
