// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
  'cloudera/search/SearchFilters',
  'cloudera/search/FilterModel',
  'knockout',
  'underscore'
], function(SearchFilters, FilterModel, ko, _) {

  describe('SearchFilters', function() {

    var $searchFilters, staticFilters, searchFilters, goodResponse,
      exportResults;

    beforeEach(function() {
      jasmine.Ajax.useMock();
      // Dummy success JSON to return from mock responses.
      goodResponse = {
        status: 200,
        responseText: JSON.stringify({success: true})
      };
      // Generate the mock data that we're sending in to the
      // search instance.
      $searchFilters = $('<div></div>').appendTo($('body'));
      staticFilters = [];
      exportResults = jasmine.createSpy('exportResults');

      searchFilters = new SearchFilters(
        $searchFilters[0], 'test/filters/url', staticFilters, exportResults);
    });

    afterEach(function() {
      $searchFilters.remove();
    });

    it('defines some attributes', function() {
      expect(searchFilters.filterContainer).toBeDefined();
      expect(searchFilters.filters).toBeDefined();
      expect(searchFilters.staticFilters).toBeDefined();
      expect(searchFilters.appliedFilters).toBeDefined();
      expect(searchFilters.filterPath).toBeDefined();
      expect(searchFilters.activeFilter).toBeDefined();
      expect(searchFilters.exportResults).toEqual(exportResults);
    });

    it('applies bindings upon instantiation', function() {
      spyOn(ko, 'applyBindings');
      searchFilters = new SearchFilters(
        $searchFilters[0], 'test/filters/url', []);
      expect(ko.applyBindings.callCount).toEqual(1);
      var args = ko.applyBindings.argsForCall[0];
      expect(args[0]).toEqual(searchFilters);
      expect(args[1]).toEqual($searchFilters[0]);
    });

    it('can get a compare type from its name', function() {
      var compareType1 = { name: 'compareType1' };
      var compareType2 = { name: 'compareType2' };
      var filter = {
        compareTypes: [compareType1, compareType2]
      };
      var compareType = searchFilters.compareTypeFromName(filter, 'compareType2');
      expect(compareType).toEqual(compareType2);
      compareType = searchFilters.compareTypeFromName(filter, 'unknown');
      expect(compareType).toBeFalsy();
    });

    it('correctly appends filters', function() {
      var filter = new FilterModel();
      expect(searchFilters.appliedFilters().length).toEqual(0);
      searchFilters.appendFilter(filter);
      expect(searchFilters.appliedFilters().length).toEqual(1);
    });

    it('does not append if any filters are duplicates', function() {
      var filter = new FilterModel();
      filter.filter('filter');
      filter.compareType('compareType');
      filter.value('value');
      expect(searchFilters.appliedFilters().length).toEqual(0);
      searchFilters.appendFilter(filter);
      expect(searchFilters.appliedFilters().length).toEqual(1);
      // Try to add exact same filter.
      var filter2 = new FilterModel();
      filter2.filter('filter');
      filter2.compareType('compareType');
      filter2.value('value');
      searchFilters.appendFilter(filter2);
      expect(searchFilters.appliedFilters().length).toEqual(1);
    });

    it('sets filter to active if not complete', function() {
      var filter = new FilterModel();
      expect(filter.complete()).toBeFalsy();
      expect(filter.active()).toBeFalsy();
      searchFilters.appendFilter(filter);
      expect(filter.active()).toBeTruthy();
    });

    it('tracks which filter is the active one and allows only one', function() {
      var filter1 = new FilterModel();
      filter1.filter('filter1');
      filter1.compareType('compareType1');
      filter1.value('value1');
      var filter2 = new FilterModel();
      filter2.filter('filter2');
      filter2.compareType('compareType2');
      filter2.value('value2');
      searchFilters.appendFilter(filter1);
      searchFilters.appendFilter(filter2);
      expect(searchFilters.appliedFilters().length).toEqual(2);
      filter1.active(true);
      expect(filter1.active()).toBeTruthy();
      expect(filter2.active()).toBeFalsy();
      filter2.active(true);
      expect(filter1.active()).toBeFalsy();
      expect(filter2.active()).toBeTruthy();
    });

    it('removes incomplete active filters when deactivated', function() {
      var filter1 = new FilterModel();
      filter1.filter('filter1');
      searchFilters.appendFilter(filter1);
      expect(filter1.active()).toBeTruthy();
      expect(searchFilters.activeFilter()).toEqual(filter1);
      expect(searchFilters.appliedFilters().length).toEqual(1);
      searchFilters.deactivateAllFilters();
      expect(searchFilters.appliedFilters().length).toEqual(0);
      expect(searchFilters.activeFilter()).toBeFalsy();
    });

    it('tracks filters and removes uncompleted deactivated ones', function() {
      spyOn(searchFilters, 'removeFilter');
      var filter1 = new FilterModel();
      searchFilters.appendFilter(filter1);
      expect(filter1.active()).toBeTruthy();
      expect(filter1.complete()).toBeFalsy();
      expect(searchFilters.appliedFilters().length).toEqual(1);
      filter1.active(false);
      expect(searchFilters.removeFilter).wasCalled();
    });

    it('fires an event when an active, completed filter is set inactive', function() {
      spyOn($, 'publish');
      var filter1 = new FilterModel();
      searchFilters.appendFilter(filter1);
      expect($.publish).wasNotCalled();
      expect(filter1.active()).toBeTruthy();
      expect(filter1.complete()).toBeFalsy();
      expect(searchFilters.appliedFilters().length).toEqual(1);
      filter1.filter('filter1');
      filter1.compareType('compareType1');
      filter1.value('value1');
      expect(filter1.complete()).toBeTruthy();
      filter1.active(false);
      expect($.publish).wasCalledWith('searchFilters.appliedFilters.changed');
    });

    it('knows how to remove a filter', function() {
      var filter1 = new FilterModel();
      filter1.filter('filter1');
      filter1.compareType('compareType1');
      filter1.value('value1');
      var filter2 = new FilterModel();
      filter2.filter('filter2');
      filter2.compareType('compareType2');
      filter2.value('value2');
      searchFilters.appendFilter(filter1);
      searchFilters.appendFilter(filter2);
      expect(searchFilters.appliedFilters().length).toEqual(2);

      // Start tracking how many times the array is changed.
      var count = 0;
      searchFilters.appliedFilters.subscribe(function() {
        count += 1;
      });
      searchFilters.removeFilter(filter1);
      expect(searchFilters.appliedFilters().length).toEqual(1);
      expect(count).toEqual(1);
      expect(filter1.active()).toBeFalsy();
      expect(searchFilters.activeFilter()).toEqual(null);
    });

    it('deactivates an active filter when it is removed', function() {
      var filter = new FilterModel();
      filter.filter('filter1');
      filter.compareType('compareType1');
      filter.value('value1');
      searchFilters.appendFilter(filter);
      filter.active(true);
      expect(searchFilters.activeFilter()).toEqual(filter);
      searchFilters.removeFilter(filter);
      expect(searchFilters.activeFilter()).toEqual(null);
    });

    it('correctly deactivates filters', function() {
      var filter = new FilterModel();
      searchFilters.appendFilter(filter);
      expect(filter.active()).toBeTruthy();
      searchFilters.deactivateAllFilters();
      expect(filter.active()).toBeFalsy();
    });

    it('loads filters via a getJSON call', function() {
      spyOn(searchFilters, 'loadFiltersHandler');
      searchFilters.loadFilters();
      var request = mostRecentAjaxRequest();
      request.response(goodResponse);
      expect(searchFilters.loadFiltersHandler).wasCalled();
    });

    it('knows how to handle the filters response from the server', function() {
      searchFilters.filters([1, 2]);
      var responseFromServer = {
        message: 'OK',
        data: [4, 5, 6]
      };
      searchFilters.loadFiltersHandler(responseFromServer);
      expect(searchFilters.filters().length).toEqual(3);
      var filters = searchFilters.filters();
      expect(filters[0]).toEqual(4);
      expect(filters[1]).toEqual(5);
      expect(filters[2]).toEqual(6);
    });

    it('calls a callback when filters are loaded', function() {
      searchFilters.filters([1, 2]);
      var callback = jasmine.createSpy('load filters callback');
      searchFilters.loadFilters(callback);
      var request = mostRecentAjaxRequest();
      request.response(goodResponse);
      expect(callback).wasCalled();
    });

    it('can return all the applied filters', function() {
      var makeTestFilterModel = function(index) {
        var model = new FilterModel();
        model.filter({ propertyName: 'propertyName ' + index});
        model.compareType({ name: 'compareType ' + index });
        model.value('value ' + index);
        return model;
      };
      var model1 = makeTestFilterModel(1);
      var model2 = makeTestFilterModel(2);
      var model3 = makeTestFilterModel(3);
      // Create a special one that has 'values' set instead of 'value'.
      var model4 = new FilterModel();
      model4.filter({ propertyName: 'propertyName 4'});
      model4.compareType({ name: 'compareType 4' });
      model4.values(['value 4.1', 'value 4.2']);

      // The integrity check in getAppliedFilters should discard
      // model3 because of its null value.
      model3.value(null);
      // Add a static filter, it should come first. We don't care about
      // the rest of the properties for the purposes of this test.
      var staticModel = {
        values: ['value 0']
      };
      searchFilters.staticFilters = [staticModel];
      searchFilters.appliedFilters([model1, model2, model3, model4]);

      var appliedFilters = searchFilters.getAppliedFilters();
      expect(appliedFilters.length).toEqual(4);
      expect(appliedFilters[0].values).toEqual(['value 0']);
      expect(appliedFilters[1].values).toEqual(['value 1']);
      expect(appliedFilters[2].values).toEqual(['value 2']);
      var values = appliedFilters[3].values;
      expect(values.length).toEqual(2);
      expect(values[0]).toEqual('value 4.1');
      expect(values[1]).toEqual('value 4.2');
    });

    it('can find filters', function() {
      var filter1 = { propertyName: 'filter1' };
      var filter2 = { propertyName: 'filter2' };
      var filter3 = { propertyName: 'filter3' };
      searchFilters.filters([filter1, filter2, filter3]);
      var foundFilter = searchFilters.findFilter('filter2');
      expect(foundFilter).toEqual(filter2);
      foundFilter = searchFilters.findFilter('unknown');
      expect(foundFilter).toEqual(null);
    });

    it('can register filters', function() {
      var filter1 = {};
      expect(searchFilters.filters().length).toEqual(0);
      searchFilters.registerFilter(filter1);
      expect(searchFilters.filters().length).toEqual(1);
      expect(searchFilters.filters()[0]).toEqual(filter1);

      var filter2 = {};
      var filter3 = {};
      searchFilters.registerAllFilters([filter2, filter3]);
      expect(searchFilters.filters().length).toEqual(3);
    });

    it('won\'t register a filter that is in the static filters list', function() {
      var filter1 = {};
      expect(searchFilters.filters().length).toEqual(0);
      searchFilters.registerFilter(filter1);
      expect(searchFilters.filters().length).toEqual(1);
      expect(searchFilters.filters()[0]).toEqual(filter1);
      
      var staticFilter = {propertyName: 'static'};
      searchFilters.staticFilters = [staticFilter];

      var filter2 = {propertyName: 'static'};
      var filter3 = {};
      searchFilters.registerAllFilters([filter2, filter3]);
      expect(searchFilters.filters().length).toEqual(2);
      expect(_.find(searchFilters.filters(), function(filter) {
        return staticFilter.propertyName === filter.propertyName;
      })).toBeUndefined();
    });

    it('adds utility methods to registered filters', function() {
      var filter = {};
      searchFilters.registerFilter(filter);
      expect(filter.has).toBeDefined();
      expect(filter.hasnt).toBeDefined();
      expect(filter.executeQuery).not.toBeDefined();

      // Test the utility methods.
      filter.catpants = true;
      expect(filter.has('catpants')).toBeTruthy();
      expect(filter.has('catpants', 'doghat')).toBeFalsy();
      expect(filter.hasnt('catpants')).toBeFalsy();
      expect(filter.hasnt('catpants', 'doghat')).toBeTruthy();
    });

    describe('executeQuery', function() {
      var filter = {
        valueUri: 'some/made/up/url'
      };
      var mockQuery;

      beforeEach(function() {
        mockQuery = {
          term: '',
          callback: jasmine.createSpy()
        };
        searchFilters.registerFilter(filter);
        expect(filter.executeQuery).toBeDefined();
      });

      it('should not make request for empty term', function() {
        filter.executeQuery(mockQuery);
        var request = mostRecentAjaxRequest();
        expect(request.url).not.toEqual(filter.valueUri);
        expect(mockQuery.callback).wasNotCalled();
      });

      it('should make request for partial term', function() {
        mockQuery.term = 'cat';
        filter.executeQuery(mockQuery);
        var request = mostRecentAjaxRequest();
        expect(request.url).toEqual(filter.valueUri + '?q=cat');
        var response = [
          'Sir Catpants',
          'Son of Catpants'
        ];
        request.response({
          status: 200,
          responseText: JSON.stringify(response)
        });
        expect(mockQuery.callback).wasCalled();
        var args = mockQuery.callback.mostRecentCall.args;
        expect(args.length).toEqual(1);
        var results = args[0].results;
        expect(results).toBeDefined();
        expect(results.length).toEqual(2);
        var result = results[0];
        expect(result.id).toEqual(response[0]);
        expect(result.text).toEqual(response[0]);
        result = results[1];
        expect(result.id).toEqual(response[1]);
        expect(result.text).toEqual(response[1]);
      });
    });

    it('can clear out the filters', function() {
      var filter1 = { propertyName: 'filter1' };
      var filter2 = { propertyName: 'filter2' };
      var filter3 = { propertyName: 'filter3' };
      searchFilters.filters([filter1, filter2, filter3]);
      expect(searchFilters.filters().length).toEqual(3);
      searchFilters.clearFilters();
      expect(searchFilters.filters().length).toEqual(0);
    });

    it('converts single value to singleton array for values', function() {
      spyOn(searchFilters, 'compareTypeFromName').andReturn('compareType');
      spyOn(searchFilters, 'appendFilter').andCallThrough();
      var result = searchFilters.applyFilter('filter', 'compareTypeName', 'catpants');
      expect(result).toBeTruthy();
      var filterModel = searchFilters.appendFilter.mostRecentCall.args[0];
      expect(filterModel.values().length).toEqual(1);
      expect(filterModel.values()[0]).toEqual('catpants');
    });

    it('will convert appended filter compare type if given string', function() {
      spyOn(searchFilters, 'compareTypeFromName').andReturn('compareType');
      spyOn(searchFilters, 'appendFilter').andReturn(true);
      var result = searchFilters.applyFilter('filter', 'compareTypeName', 'value');
      expect(result).toBeTruthy();
      expect(searchFilters.compareTypeFromName).wasCalled();

      searchFilters.compareTypeFromName.reset();
      result = searchFilters.applyFilter('filter', {}, 'value');
      expect(result).toBeTruthy();
      expect(searchFilters.compareTypeFromName).wasNotCalled();
    });

    it('will fail if compare type is not found', function() {
      spyOn(searchFilters, 'compareTypeFromName').andReturn(null);
      spyOn(searchFilters, 'appendFilter').andReturn(true);
      var result = searchFilters.applyFilter('filter', 'compareTypeName', 'value');
      expect(result).toBeFalsy();
      expect(searchFilters.compareTypeFromName).wasCalled();
    });

    it('can add another filter while one is active', function() {
      spyOn(searchFilters, 'deactivateAllFilters');
      spyOn(searchFilters, 'appendFilter');
      searchFilters.addAnother();
      expect(searchFilters.deactivateAllFilters).wasCalled();
      expect(searchFilters.appendFilter).wasCalled();
    });
  });
});